using System;
using System.Threading;
using SDL2;
using Serilog;
using Serilog.Core;
using SlatedGameToolkit.Framework.Exceptions;
using SlatedGameToolkit.Framework.StateSystem;

namespace SlatedGameToolkit.Framework {
    /// <summary>
    /// The main engine that will host the game loop.
    /// </summary>
    public static class GameEngine {
        public static readonly Logger logger = new LoggerConfiguration().
        WriteTo.Console().
        WriteTo.File("SlatedGameToolKit.Framework.Log").
        CreateLogger();
        private static readonly object ignitionLock = new object();
        private static Thread thread;
        private static volatile bool exit = false, loopCompleted = true;
        private static long updateDeltaTime = 50, frameDeltaTime = 0;

        /// <summary>
        /// The amount of updates per second.
        /// Is floored to milleseconds.
        /// </summary>
        /// <value>The updates per second.</value>
        public static double UpdatesPerSecond {
            get {
                return TimeSpan.FromMilliseconds(1 / updateDeltaTime).TotalSeconds;
            }
            set {
                Interlocked.Exchange(ref updateDeltaTime, (long) TimeSpan.FromSeconds(1 / value).TotalMilliseconds);
            }
        }

        /// <summary>
        /// The target frames per second. Will not go above this number, but may dip below this number.
        /// This value is floored to millesconds.
        /// </summary>
        /// <value>The target frames per second.</value>
        public static double targetFPS {
            get {
                return TimeSpan.FromMilliseconds(1 / frameDeltaTime).TotalSeconds;
            }
            set {
                Interlocked.Exchange(ref frameDeltaTime, (long) TimeSpan.FromSeconds(1 / value).TotalMilliseconds);
            }
        }

        private static void Loop(Object o) {
            if (!(o is Manager)) throw new InternalFrameworkException(String.Format("Expected manager object for asynchronous loop. Got {0}", o));
            Manager manager = (Manager) o;
            long currentTime = DateTimeOffset.Now.ToUnixTimeMilliseconds();
            long timePassedFromLastUpdate = 0;
            long timePassedFromLastRender = 0;
            loopCompleted = false;
            while (!exit) {
                long updateDeltaTime = Interlocked.Read(ref GameEngine.updateDeltaTime);
                long frameDeltaTime = Interlocked.Read(ref GameEngine.frameDeltaTime);
                long frameStart = DateTimeOffset.Now.ToUnixTimeMilliseconds();
                long difference = frameStart - currentTime;
                currentTime = frameStart;
                
                timePassedFromLastUpdate += difference;
                while (timePassedFromLastUpdate > updateDeltaTime) {
                    manager.update(updateDeltaTime);
                    timePassedFromLastUpdate -= updateDeltaTime;
                }

                timePassedFromLastRender += difference;
                if (timePassedFromLastRender > frameDeltaTime) {
                    manager.render(timePassedFromLastUpdate / updateDeltaTime);
                    timePassedFromLastRender = 0;
                }
            }
            loopCompleted = true;
            SDL.SDL_Quit();
        }

        /// <summary>
        /// Requests to stop the game engine.
        /// Game engine will finish the frame its currently in.
        /// This needs to be called to perform proper clean up.
        /// </summary>
        /// <returns>False if the stop request has already been made.</returns>
        public static bool Stop() {
            lock (ignitionLock) {
                if (exit) return false;
                exit = true;
                return true;
            }
        }

        /// <summary>
        /// Requests to start the game engine. 
        /// </summary>
        /// <returns>True iff the engine is not already running.</returns>
        private static bool Ignite(Manager manager) {
            if (manager == null) throw new ArgumentNullException("manager");
            SDL.SDL_version SDLVersion;
            SDL.SDL_version SDLBuiltVersion;
            SDL.SDL_GetVersion(out SDLVersion);
            SDL.SDL_VERSION(out SDLBuiltVersion);
            logger.Information(String.Format("Attempting to initiate game engine with SDL version: {0}", SDLVersion.ToString()));
            if (SDL.SDL_VERSION_ATLEAST(SDLBuiltVersion.major, SDLBuiltVersion.minor, SDLBuiltVersion.patch)) {
                logger.Warning(String.Format("Engine was designed with SDL version {0}, currently running on {1}", SDLVersion.ToString(), SDLBuiltVersion.ToString()));
                if (SDLVersion.major < 2) {
                    logger.Error("This engine was designed to work with SDL2. The version you're currently running is severely outdated.");
                    throw new FrameworkUsageException("Outdated SDL binaries.");
                }
            }
            lock (ignitionLock) {
                if (!loopCompleted) return false;
                loopCompleted = false;

                if (SDL.SDL_GL_SetAttribute(SDL.SDL_GLattr.SDL_GL_CONTEXT_MAJOR_VERSION, 3) < 0) throw new FrameworkSDLException();
                if (SDL.SDL_GL_SetAttribute(SDL.SDL_GLattr.SDL_GL_CONTEXT_MINOR_VERSION, 1) < 0) throw new FrameworkSDLException();
                if (SDL.SDL_GL_SetAttribute(SDL.SDL_GLattr.SDL_GL_CONTEXT_PROFILE_MASK, SDL.SDL_GLprofile.SDL_GL_CONTEXT_PROFILE_CORE) < 0) throw new FrameworkSDLException();

                exit = false;
                if (SDL.SDL_Init(SDL.SDL_INIT_VIDEO) != 0) {
                    throw new FrameworkSDLException();
                }
                if (SDL.SDL_Init(SDL.SDL_INIT_AUDIO) != 0) {
                    throw new FrameworkSDLException();
                }
                thread = new Thread(Loop);
                thread.Start(manager);
                return true;
            }
        }
    }
}